from tenant.forms import ModelForm
from .models import Sequence


class SequenceCreateForm(ModelForm):
    class Meta:
        model = Sequence
        fields = [
            'name',
            'nextval',
            'step'
        ]



import logging
from django import template
from django.conf import settings
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe

_logger = logging.getLogger(__name__)
register = template.Library()


@register.simple_tag(name='combine_templates')
def combine_templates(template_name):
    html = ''
    for app in settings.INSTALLED_APPS:
        try:
            html += render_to_string(f'{app.split(".")[-1]}/{template_name}')
        except template.TemplateDoesNotExist:
            continue
    return mark_safe(html)


@register.filter(name='get_attr')
def get_attr_from_string(param, value):
    try:
        attr = getattr(param, value)
    except AttributeError:
        _logger.exception(f'Object {param} has no attribute {value}')
        return ''
    if callable(attr):
        return attr()
    else:
        return attr


@register.filter(name='render_query_params')
def query_params_to_html(params):
    html = ''
    for param in params:
        html += build(param)
    return mark_safe(html)


def build(param):
    if not param.get('params'):
        return build_param(param)
    else:
        return build_params(param)


def get_value(param):
    if param.get('data_type', '') == 'bool':
        return f'data-value={param["value"]}'
    else:
        return f'data-value=""'


def build_params(param):
    start = f'<div class="query-param" tabindex="-1" data-param="{param["param"]}" ' \
            f'data-param-label="{param["label"]}">' \
            f'<p class="px-1 arrow">{param["label"]}</p>' \
            f'<div class="query-params is-hidden" data-param="{param["param"]}">'
    stop = '</div></div>'
    params = []
    for p in param.get('params'):
        params.append(build(p))
    return f'{start}{"".join(params)}{stop}'


def build_param(param):
    return f'<div class="query-param" tabindex="-1" data-type="{param["data_type"]}" ' \
           f'data-param-invert="{"true" if param.get("invert") else "false"}" ' \
           f'data-param="{param["param"]}" data-param-label="{param["label"]}" ' \
           f'{get_value(param)} data-step="{param.get("step", "1")}">' \
           f'<p class="px-1 arrowless">{param["label"]}</p>' \
           f'<div class="param-options is-hidden">{build_options(param)}</div></div>'


def build_options(param):
    options = ''
    for choice in param.get('choices', []):
        options += f'<option value="{choice[0]}">{choice[1]}</option>'
    return options



import datetime
import logging
import json
from dataclasses import dataclass, field
from django.db.models.functions import Lower
from django.db.models import Model, QuerySet, Q, CharField, Manager
from django.utils.translation import gettext_lazy as _
from django.core.exceptions import FieldDoesNotExist
from django.core.paginator import Paginator
from accrete.contrib.ui.filter import Filter

_logger = logging.getLogger(__name__)

DEFAULT_PAGINATE_BY = 40


@dataclass
class ClientAction:

    name: str
    url: str = ''
    query_params: str = ''
    attrs: list[str] = field(default_factory=list)
    submit: bool = False
    form_id: str = 'form'
    class_list: list = field(default_factory=list)


@dataclass
class BreadCrumb:

    name: str
    url: str


@dataclass
class TableField:

    label: str
    name: str


@dataclass
class ListContext:

    model: type[Model]
    get_params: dict
    title: str = None
    context: dict = field(default_factory=dict)
    queryset: QuerySet = None
    extra_query: Q = None
    related_fields: list[str] = field(default_factory=list)
    prefetch_fields: list[str] = field(default_factory=list)
    paginate_by: int = DEFAULT_PAGINATE_BY
    order_by: list[str] = None
    column_width: int = 12
    filter_relation_depth: int = 4
    default_filter_term: str = ''
    actions: list[ClientAction] = field(default_factory=list)
    breadcrumbs: list[BreadCrumb] = field(default_factory=list)
    obj_label: str = None
    fields: list[TableField] = field(default_factory=list)

    def get_queryset(self):
        order = self.get_order()
        if isinstance(order, str):
            if order == 'None':
                order = None
        return (self.queryset or queryset_from_querystring(
            self.model, self.get_params.get('q', '[]'), order)
                .filter(self.extra_query or Q())
                .select_related(*self.related_fields)
                .prefetch_related(*self.prefetch_fields)
                .distinct())

    def get_page_number(self, paginator):
        page_number = self.get_params.get('page', '1')

        try:
            page_number = int(page_number)
        except ValueError:
            page_number = 1

        if page_number < 1:
            page_number = 1
        elif page_number > paginator.num_pages:
            page_number = paginator.num_pages
        return page_number

    def get_order(self):
        return self.get_params.get('order_by', None) or self.order_by

    def get_paginate_by(self):
        paginate_by = self.get_params.get('paginate_by', self.paginate_by)
        try:
            paginate_by = int(paginate_by)
        except ValueError:
            paginate_by = self.paginate_by
        return paginate_by

    def get_context(self):
        paginate_by = self.get_paginate_by()
        paginator = Paginator(self.get_queryset(), paginate_by)
        page = paginator.page(self.get_page_number(paginator))
        context = {
            'paginate_by': paginate_by,
            'order_by': self.get_params.get('order_by', self.model._meta.ordering),
            'paginator': paginator,
            'page': page,
            'list_pagination': True,
            'column_width': self.column_width,
            'title': self.title or self.model._meta.verbose_name_plural,
            'filter_terms': Filter(self.model, self.filter_relation_depth).get_query_terms(),
            'default_filter_term': self.default_filter_term,
            'view_type': 'list',
            'breadcrumbs': self.breadcrumbs,
            'querystring': build_querystring(self.get_params),
            'actions': self.actions,
            'obj_label': self.obj_label or _('Name'),
            'fields': self.fields
        }
        context.update(self.context)
        return context


@dataclass
class DetailContext:

    obj: Model
    get_params: dict
    order_by: str = None
    paginate_by: int = DEFAULT_PAGINATE_BY
    title: str = None
    queryset: type[QuerySet] = None
    extra_query: Q = None
    related_fields: list[str] = field(default_factory=list)
    prefetch_fields: list[str] = field(default_factory=list)
    actions: list[ClientAction] = field(default_factory=list)
    breadcrumbs: list[BreadCrumb] = field(default_factory=list),
    context: dict = field(default_factory=dict)

    def get_queryset(self):
        order = self.get_order()
        if isinstance(order, str):
            if order == 'None':
                order = None
        return (self.queryset or queryset_from_querystring(
            self.obj._meta.model, self.get_params.get('q', '[]'), order)
                .filter(self.extra_query or Q())
                .select_related(*self.related_fields)
                .prefetch_related(*self.prefetch_fields)
                .distinct())

    def get_order(self):
        return self.get_params.get('order_by', None) or self.order_by

    def get_paginate_by(self):
        paginate_by = self.get_params.get('paginate_by', self.paginate_by)
        try:
            paginate_by = int(paginate_by)
        except ValueError:
            paginate_by = self.paginate_by
        return paginate_by

    def get_pagination_context(self):
        if not hasattr(self.obj, 'get_absolute_url'):
            _logger.warning(
                'Detail pagination disabled for models without the '
                'get_absolute_url attribute. Set paginate_by to 0 to '
                'deactivate pagination.'
            )
            return {}
        queryset = self.get_queryset()
        idx = (*queryset,).index(self.obj)
        previous_object_url = (
            queryset[idx - 1] if idx - 1 >= 0 else queryset.last()
        ).get_absolute_url()
        next_object_url = (
            queryset[idx + 1] if idx + 1 <= queryset.count() - 1 else queryset.first()
        ).get_absolute_url()
        ctx = {
            'previous_object_url': previous_object_url,
            'next_object_url': next_object_url,
            'current_object_idx': idx + 1,
            'total_objects': queryset.count(),
            'detail_pagination': True
        }
        return ctx

    def get_context(self):
        paginate_by = self.get_paginate_by()
        ctx = {
            'object': self.obj,
            'title': self.title or self.obj,
            'order_by': self.get_params.get('order_by', self.obj._meta.model._meta.ordering),
            'paginate_by': paginate_by,
            'detail_pagination': False,
            'view_type': 'detail',
            'breadcrumbs': self.breadcrumbs,
            'querystring': build_querystring(self.get_params, ['page']),
            'actions': self.actions
        }
        if self.paginate_by > 0:
            ctx.update(self.get_pagination_context())
        ctx.update(self.context)
        return ctx


@dataclass
class FormContext:

    model: Model | type[Model]
    get_params: dict
    title: str = None
    context: dict = field(default_factory=dict)
    form_id: str = 'form'
    add_default_actions: bool = True
    discard_url: str = None
    actions: list[ClientAction] = field(default_factory=list)

    def get_default_form_actions(self):
        actions = [
            ClientAction(
                name=_('Save'),
                submit=True,
                class_list=['is-success'],
                form_id=self.form_id
            )
        ]
        try:
            url = self.discard_url or self.model.get_absolute_url()
        except TypeError:
            raise TypeError(
                'Supply the discard_url parameter if FormContext is called '
                'with a model class instead of an instance.'
            )
        except AttributeError as e:
            _logger.error(
                'Supply the discard_url parameter if FormContext is '
                'called with a model instance that has the get_absolute_url '
                'method not defined.'
            )
            raise e

        actions.append(
            ClientAction(
                name=_('Discard'),
                url=url,
            )
        )
        return actions

    def get_title(self):
        if self.title:
            return self.title
        try:
            int(self.model.pk)
            return f'Edit {self.model}'
        except TypeError:
            return f'Add {self.model._meta.verbose_name}'

    def get_context(self):
        ctx = {
            'title': self.get_title(),
            'view_type': 'form',
            'form_id': self.form_id,
            'querystring': build_querystring(self.get_params, ['page']),
            'actions': []
        }
        if self.add_default_actions:
            ctx.update({'actions': self.get_default_form_actions()})
        ctx['actions'].extend(self.actions)
        ctx.update(self.context)
        return ctx


def build_querystring(get_params: dict, extra_params: list[str] = None) -> str:
    querystring = f'?q={get_params.get("q", "[]")}'
    if paginate_by := get_params.get('paginate_by', False):
        querystring += f'&paginate_by={paginate_by}'
    # if page := get_params.get('page', False):
    #     querystring += f'&page={page}'
    if order_by := get_params.get('order_by', False):
        querystring += f'&order_by={order_by}'
    for param in extra_params or []:
        if value := get_params.get(param, False):
            querystring += f'&{param}={value}'
    return querystring


def queryset_from_querystring(
        model: type[Model|Manager],
        query_string: str,
        order_by: str|list[str] = None,
) -> QuerySet:
    """
    param url_query_string: json serializable string
    [[{"name__iexact": "asdf"}, {"active": true}], [{"group__name__isnull": false}]]
    """

    if not isinstance(model, Manager):
        model = model.objects

    query_data = json.loads(query_string)
    query = Q()

    for query_block in query_data:
        block_query = Q()
        for query_term in query_block:
            if isinstance(query_term, dict):
                for param, value in query_term.items():
                    block_query &= get_query(model.model, param, value)
            elif isinstance(query_term, list):
                inner_block_query = Q()
                for inner_query_term in query_term:
                    for param, value in inner_query_term.items():
                        inner_block_query |= get_query(model.model, param, value)
                block_query &= inner_block_query
        query |= block_query

    queryset = model.filter(query)

    if order_by is None:
        return queryset

    try:
        order = order_by.split(',')
    except AttributeError:
        order = order_by or []

    for o in order_by:
        desc = False
        if o.startswith('-'):
            o = o[1:]
            desc = True
        try:
            model_field = model.model._meta.get_field(o)
        except FieldDoesNotExist:
            model_field = None
            if o == 'pk':
                pass
        if isinstance(model_field, CharField):
            if desc:
                order.append(Lower(o).desc())
            else:
                order.append(Lower(o).asc())
        else:
            order.append(f'{"-" if desc else ""}{o}')
    return queryset.order_by(*order)


def get_related_model(model, path):
    related_model = model
    for part in path:
        try:
            related_model = related_model._meta.fields_map[part].related_model
        except (AttributeError, KeyError):
            continue
    return related_model


def get_query(model, param: str, value) -> Q:
    invert = False
    parts = param.split('__')
    if param.startswith('~'):
        param = param[1:]
        invert = True
    if len(parts) == 1:
        parts.append('exact')
    attribute = parts[-2]

    if not attribute.startswith('_c_'):
        query = Q(**{param: value})
        return ~query if invert else query

    operator = parts[-1]
    related_model = get_related_model(model, parts[:-2])
    func = getattr(related_model, attribute[3:])

    obj_ids = [
        obj.id for obj in
        filter(
            lambda instance:
            evaluate(instance, func, value, operator),
            related_model.objects.all()
        )
    ]

    query = Q(**{
        f'{"__".join(parts[:-2])}{"__" if parts[:-2] else ""}id__in': obj_ids
    })
    return ~query if invert else query


def evaluate(instance, func, value, operator):
    return_type = func.__annotations__.get('return', str)
    try:
        value = return_type(value)
    except TypeError as e:
        if isinstance(return_type, datetime.datetime):
            value = datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M')
        elif isinstance(return_type, datetime.date):
            value = datetime.datetime.strptime(value, '%Y-%m-%d')
        else:
            raise e

    return_value = func(instance)

    if operator == 'exact':
        return value == return_value
    elif operator == 'icontains':
        return value.lower() in return_value.lower()
    elif operator == 'contains':
        return value in return_value
    elif operator == 'gt':
        return return_value > value
    elif operator == 'gte':
        return return_value >= value
    elif operator == 'lte':
        return return_value <= value
    elif operator == 'lt':
        return return_value < value



from django.urls import path

from . import views

app_name = 'user'

urlpatterns = [
    path('login/', views.LoginView.as_view(), name='login'),
    path('logout/', views.LogoutView.as_view(), name='logout')
]



from .tenant import get_tenant, set_tenant, get_member, set_member, Unscoped
from .decorators import tenant_required
from .views import TenantRequiredMixin
from .forms import save_form


def unscoped():
    return Unscoped(get_tenant())



from django.db import models
from django.utils.translation import gettext_lazy as _

from accrete.models import TenantModel


class Sequence(TenantModel):

    class Meta:
        db_table = 'accrete_sequence'
        constraints = [
            models.UniqueConstraint(
                name='unique_name_per_tenant',
                fields=['name', 'tenant']
            ),
            models.UniqueConstraint(
                name='unique_val_per_name_tenant',
                fields=['name', 'tenant', 'nextval']
            )
        ]

    name = models.CharField(
        verbose_name=_('Name'),
        max_length=50,
    )

    nextval = models.PositiveBigIntegerField(
        verbose_name=_('Next Value'),
        default=1
    )

    step = models.PositiveIntegerField(
        verbose_name=_('Step'),
        default=1
    )

    def __str__(self):
        return self.name



from functools import wraps

from django.shortcuts import redirect
from django.contrib.auth.views import login_required
from . import config


def tenant_required(
        redirect_field_name: str = None,
        login_url: str = None
):
    def decorator(f):
        @wraps(f)
        @login_required(
            redirect_field_name=redirect_field_name,
            login_url=login_url
        )
        def _wrapped_view(request, *args, **kwargs):
            tenant = request.tenant
            if not tenant:
                return redirect(config.ACCRETE_TENANT_NOT_SET_URL)
            return f(request, *args, **kwargs)
        return _wrapped_view
    return decorator



from .filter import Filter
from .helper import (
    ClientAction,
    DetailContext,
    FormContext,
    ListContext,
    BreadCrumb,
    TableField,
    build_querystring,
    queryset_from_querystring
)



from django.test import TestCase
from django.contrib.auth.models import AnonymousUser
from django.contrib.auth import get_user_model
from django.forms import ValidationError
from django.db.models import ObjectDoesNotExist
from django.views import View
from django.test import Client
from django.test.client import RequestFactory
from django.conf import settings

from accrete import models
from accrete import forms
from .views.mixins import TenantRequiredMixin

User = get_user_model()


class FormTestCase(TestCase):
    def setUp(self):
        self.tenant = models.Tenant(name='Test Tenant')
        self.tenant.save()
        self.user = User(email='test@user.test', username='Test User')
        self.user.save()

    def test_form_set_tenant(self):
        f = forms.Form(tenant=self.tenant.pk, user=self.user.pk)
        self.assertEqual(self.tenant, f.tenant)

    def test_form_set_user(self):
        f = forms.Form(tenant=self.tenant.pk, user=self.user.pk)
        self.assertEqual(self.user, f.user)

    def test_form_tenant_queryset(self):
        f = forms.Form(tenant=self.tenant.pk, user=self.user.pk)
        self.assertEqual(
            list(f.fields['tenant'].queryset),
            list(models.Tenant.objects.filter(pk=self.tenant.pk))
        )

    def test_from_clean_staff(self):
        self.user.is_staff = True
        self.user.save()
        f = forms.Form({
            'tenant': self.tenant
        }, tenant=self.tenant.pk, user=self.user.pk)
        self.assertTrue(f.is_valid())

    def test_form_no_user(self):
        with self.assertRaises(ObjectDoesNotExist):
            forms.Form({'tenant': self.tenant}, tenant=self.tenant.pk)

    def test_form_no_tenant(self):
        with self.assertRaises(ObjectDoesNotExist):
            forms.Form({'tenant': self.tenant}, user=self.user.pk)

    def test_form_not_a_member(self):
        with self.assertRaises(ValidationError):
            f = forms.Form(
                {'tenant': self.tenant},
                tenant=self.tenant.pk, user=self.user.pk
            )
            f.is_valid()
            f.clean()

    def test_form_with_member(self):
        member = models.Member(
            tenant=self.tenant,
            user=self.user
        )
        member.save()
        f = forms.Form(
            {'tenant': self.tenant},
            tenant=self.tenant.pk, user=self.user.pk
        )
        f.is_valid()
        f.clean()

    def test_modelform_set_tenant(self):
        f = forms.Form(tenant=self.tenant.pk, user=self.user.pk)
        self.assertEqual(self.tenant, f.tenant)

    def test_modelform_set_user(self):
        f = forms.Form(tenant=self.tenant.pk, user=self.user.pk)
        self.assertEqual(self.user, f.user)

    def test_modelform_tenant_queryset(self):
        f = forms.Form(tenant=self.tenant.pk, user=self.user.pk)
        self.assertEqual(
            list(f.fields['tenant'].queryset),
            list(models.Tenant.objects.filter(pk=self.tenant.pk))
        )

    def test_modelfrom_clean_staff(self):
        self.user.is_staff = True
        self.user.save()
        f = forms.Form({
            'tenant': self.tenant
        }, tenant=self.tenant.pk, user=self.user.pk)
        self.assertTrue(f.is_valid())

    def test_modelform_no_user(self):
        with self.assertRaises(ObjectDoesNotExist):
            forms.Form({'tenant': self.tenant}, tenant=self.tenant.pk)

    def test_modelform_no_tenant(self):
        with self.assertRaises(ObjectDoesNotExist):
            forms.Form({'tenant': self.tenant}, user=self.user.pk)

    def test_modelform_not_a_member(self):
        with self.assertRaises(ValidationError):
            f = forms.Form(
                {'tenant': self.tenant},
                tenant=self.tenant.pk, user=self.user.pk
            )
            f.is_valid()
            f.clean()

    def test_modelform_with_member(self):
        member = models.Member(
            tenant=self.tenant,
            user=self.user
        )
        member.save()
        f = forms.Form(
            {'tenant': self.tenant},
            tenant=self.tenant.pk, user=self.user.pk
        )
        f.is_valid()
        f.clean()


class MiddlewareTestCase(TestCase):
    def setUp(self):
        self.user = User(email='test@user.test', username='Test User')
        self.user.save()

    def test_no_tenant(self):
        c = Client()
        c.force_login(self.user)
        res = c.get('/')
        self.assertFalse(c.cookies.get('tenant_id'))
        self.assertFalse(res.wsgi_request.tenant)

    def test_get_non_existing(self):
        c = Client()
        c.force_login(self.user)
        res = c.get('/', {'tenant_id': 1})
        self.assertFalse(c.cookies.get('tenant_id'))
        self.assertFalse(res.wsgi_request.tenant)
    
    def test_get_no_member(self):
        tenant = models.Tenant(name='Test Tenant')
        tenant.save()
        c = Client()
        c.force_login(self.user)
        res = c.get('/', {'tenant_id': tenant.pk + 1})
        with self.assertRaises(KeyError):
            cookie = c.cookies['tenant_id']
        self.assertFalse(res.wsgi_request.tenant)
        self.assertFalse(res.wsgi_request.member)

    def test_get(self):
        tenant = models.Tenant(name='Test Tenant')
        tenant.save()
        member = models.Member(
            tenant=tenant,
            user=self.user
        )
        member.save()
        c = Client()
        c.force_login(self.user)
        res = c.get('/')
        self.assertEqual(c.cookies['tenant_id'].value, str(tenant.pk))
        self.assertEqual(res.wsgi_request.tenant, tenant)
        self.assertEqual(res.wsgi_request.member, member)

    def test_post_no_member(self):
        c = Client()
        c.force_login(self.user)
        c.post('/', {'tenant_id': 1})
        self.assertFalse(c.cookies.get('tenant_id'))

    def test_post(self):
        tenant = models.Tenant(name='Test Tenant')
        tenant.save()
        member = models.Member(
            tenant=tenant,
            user=self.user
        )
        member.save()
        c = Client()
        c.force_login(self.user)
        c.post('/', {'tenant_id': tenant.pk})
        self.assertEqual(c.cookies['tenant_id'].value, str(tenant.pk))


class TenantViewTestCase(TestCase):
    class TestView(TenantRequiredMixin, View):
        tenant_url = '/tenant/not/set/'

    def setUp(self):
        self.factory = RequestFactory()
        self.user_without_tenant = User(
            email='test@test.com', username='Test User'
        )
        self.user_without_tenant.save()

    def test_not_logged_in(self):
        request = self.factory.get('/')
        request.user = AnonymousUser()
        view = self.TestView.as_view()
        res = view(request)
        self.assertEqual(res.status_code, 302)
        self.assertEqual(
            res.url.split('?')[0].rstrip('/'),
            settings.LOGIN_URL.rstrip('/')
        )

    def test_no_tenant(self):
        request = self.factory.get('/')
        request.user = self.user_without_tenant
        request.tenant = None
        view = self.TestView.as_view()
        res = view(request)
        self.assertEqual(res.url, '/tenant/not/set/')



from django.views.generic import View
from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth import logout
from django.contrib.auth import views
from django.shortcuts import redirect


class LoginView(views.LoginView):
    form_class = AuthenticationForm
    template_name = 'user/login.html'
    redirect_authenticated_user = True

    def form_invalid(self, form):
        user = form.get_user()
        if user is not None and not user.is_active:
            ctx = {'to_confirm': True}
            if self.extra_context:
                self.extra_context.update(ctx)
            else:
                self.extra_context = ctx
        return super().form_invalid(form)


class LogoutView(View):
    @staticmethod
    def get(request, *args, **kwargs):
        logout(request)
        return redirect('user:login')



from django.db import models
from django.contrib.auth.models import (
    AbstractBaseUser,
    PermissionsMixin
)
from django.contrib.auth.validators import UnicodeUsernameValidator
from django.contrib.auth.models import BaseUserManager
from django.utils.translation import gettext_lazy as _
from django.utils import timezone


class UserManager(BaseUserManager):
    use_in_migrations = True

    def _create_user(self, email, password, username=None, **extra_fields):
        if not email:
            raise ValueError('The email must be set')

        email = self.normalize_email(email)
        username = self.model.normalize_username(username)
        user = self.model(email=email, username=username, **extra_fields)
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, email, password=None, username=None, **extra_fields):
        extra_fields.setdefault('is_staff', False)
        extra_fields.setdefault('is_superuser', False)
        return self._create_user(email, password, username, **extra_fields)

    def create_superuser(self, email, password, username=None, **extra_fields):
        extra_fields.setdefault('is_staff', True)
        extra_fields.setdefault('is_superuser', True)
        extra_fields.setdefault('is_active', True)

        if extra_fields.get('is_staff') is not True:
            raise ValueError('Superuser must have is_staff=True.')
        if extra_fields.get('is_superuser') is not True:
            raise ValueError('Superuser must have is_superuser=True.')

        return self._create_user(email, password, username, **extra_fields)


class User(AbstractBaseUser, PermissionsMixin):
    class Meta:
        db_table = 'accrete_user'
        verbose_name = _('User')
        verbose_name_plural = _('Users')

    username_validator = UnicodeUsernameValidator()

    username = models.CharField(
        verbose_name=_('Username'),
        max_length=150,
        help_text=_(
            '150 characters or fewer.'
            'Letters, digits and @/./+/-/_ only.'
        ),
        blank=True,
        null=True,
        validators=[username_validator],
    )

    first_name = models.CharField(
        verbose_name=_('First Name'),
        max_length=150,
        blank=True,
        null=True
    )

    last_name = models.CharField(
        verbose_name=_('Last Name'),
        max_length=150,
        blank=True,
        null=True
    )

    email = models.EmailField(
        verbose_name=_('Email Address'),
        unique=True
    )

    is_staff = models.BooleanField(
        verbose_name=_('Staff Status'),
        default=False,
        help_text=_(
            'Designates whether the user can log into the admin site.'
        ),
    )

    is_active = models.BooleanField(
        verbose_name=_('Active'),
        default=True,
        help_text=_(
            'Designates whether this user should be treated as active.\n'
            'Unselect this instead of deleting accounts.'
        ),
    )

    date_joined = models.DateTimeField(
        verbose_name=_('Date Joined'),
        default=timezone.now
    )

    language_code = models.CharField(
        verbose_name=_('Language'),
        max_length=10,
        null=True,
        blank=True
    )

    objects = UserManager()

    EMAIL_FIELD = 'email'
    USERNAME_FIELD = 'email'

    def __str__(self):
        return self.username or self.email



import contextvars
import logging
from django.apps import apps

_logger = logging.getLogger(__name__)

tenant_contextvar = contextvars.ContextVar('tenant', default=None)
member_contextvar = contextvars.ContextVar('member', default=None)


def set_tenant(tenant):
    tenant_contextvar.set(tenant)


def get_tenant():
    return tenant_contextvar.get()


def set_member(member):
    tenant = member.tenant if member is not None else None
    member_contextvar.set(member)
    tenant_contextvar.set(tenant)


def get_member():
    return member_contextvar.get()


def tenant_iterator():
    for tenant in apps.get_model('accrete', 'Tenant').objects.all():
        set_tenant(tenant)
        yield tenant


class Unscoped:

    def __init__(self, tenant):
        self.tenant = tenant

    def __enter__(self):
        if self.tenant is None:
            _logger.warning(
                'Entering unscoped context manager with tenant already set to None!'
            )
        set_tenant(None)

    def __exit__(self, exc_type, exc_val, exc_tb):
        set_tenant(self.tenant)
